LANGUAGES_DATE_STRING = {
    "bg": "d MMM yyyy",  # Bulgarian
    "cs": "d. MMM yyyy",  # Czech
    "de": "d. MMM yyyy",  # German
    "en": "MMM. d, yyyy",  # English
    "es": "d MMM yyyy",  # Spanish
    "eu": "yyyy MMM d",  # Basque
    "fr": "d MMM yyyy",  # French
    "gl": "d MMM yyyy",  # Galician
    "it": "d MMM yyyy",  # Italian
    "nb": "d. MMM yyyy",  # Norwegian Bokmål
    "nl": "d MMM. yyyy",  # Dutch
    "pl": "d MMM yyyy",  # Polish
    "pt": "d MMM yyyy",  # Portuguese
    "ru": "d MMM yyyy",  # Russian
}

SUPPORTED_LANGUAGES = list(LANGUAGES_DATE_STRING.keys())
